% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/textstat_lexdiv.R
\name{textstat_lexdiv}
\alias{textstat_lexdiv}
\title{Calculate lexical diversity}
\usage{
textstat_lexdiv(x, measure = c("all", "TTR", "C", "R", "CTTR", "U", "S",
  "K", "D", "Vm", "Maas"), log.base = 10)
}
\arguments{
\item{x}{an input \link{dfm}}

\item{measure}{a character vector defining the measure to calculate.}

\item{log.base}{a numeric value defining the base of the logarithm (for
measures using logs)}

\item{...}{not used}
}
\value{
\code{textstat_lexdiv} returns a data.frame of documents and
  their lexical diversity scores.
}
\description{
Calculate the lexical diversity or complexity of text(s).
}
\details{
\code{textstat_lexdiv} calculates a variety of proposed indices for
  lexical diversity. In the following formulas, \eqn{N} refers to the total
  number of tokens, \eqn{V} to the number of types, and \eqn{f_v(i, N)} to the
  numbers of types occurring \eqn{i} times in a sample of length \eqn{N}.
  \describe{
     
  \item{\code{"TTR"}:}{The ordinary \emph{Type-Token Ratio}: \deqn{TTR =
  \frac{V}{N}}{TTR =  V / N}}
  
  \item{\code{"C"}:}{Herdan's \emph{C} (Herdan, 1960, as cited in Tweedie & 
  Baayen, 1998; sometimes referred to as \emph{LogTTR}): \deqn{C = 
  \frac{\log{V}}{\log{N}}}{C = log(V) / log(N)}}
  
  \item{\code{"R"}:}{Guiraud's \emph{Root TTR} (Guiraud, 1954, as cited in 
  Tweedie & Baayen, 1998): \deqn{R = \frac{V}{\sqrt{N}}}{R = V / sqrt(N)}}
  
  \item{\code{"CTTR"}:}{Carroll's \emph{Corrected TTR}: \deqn{CTTR = 
  \frac{V}{\sqrt{2N}}}{CTTR = V / sqrt(2N)}}
  
  \item{\code{"U"}:}{Dugast's \emph{Uber Index}  (Dugast, 1978, as cited in 
  Tweedie & Baayen, 1998): \deqn{U = \frac{(\log{N})^2}{\log{N} - \log{V}}}{U
  = log(N)^2 / log(N) - log(V)}}
  
  \item{\code{"S"}:}{Summer's index: \deqn{S = 
  \frac{\log{\log{V}}}{\log{\log{N}}}}{S = log(log(V)) / log(log(N))}}
  
  \item{\code{"K"}:}{Yule's \emph{K}  (Yule, 1944, as presented in Tweedie &
  Baayen, 1998, Eq. 16) is calculated by: \deqn{K = 10^4 \times
  \left[ -\frac{1}{N} + \sum_{i=1}^{V} f_v(i, N) \left( \frac{i}{N} \right)^2 \right] }}
  
  \item{\code{"D"}:}{Simpson's \emph{D}  (Simpson 1949, as presented in
  Tweedie & Baayen, 1998, Eq. 17) is calculated by:
  \deqn{D = \sum_{i=1}^{V} f_v(i, N) \frac{i}{N} \frac{i-1}{N-1}}}

  \item{\code{"Vm"}:}{Herdan's \eqn{V_m}  (Herdan 1955, as presented in
  Tweedie & Baayen, 1998, Eq. 18) is calculated by:
  \deqn{V_m = \sqrt{ \sum_{i=1}^{V} f_v(i, N) (i/N)^2 - \frac{i}{V} }}}

  \item{\code{"Maas"}:}{Maas' indices (\eqn{a}, \eqn{\log{V_0}} & 
  \eqn{\log{}_{e}{V_0}}): \deqn{a^2 = \frac{\log{N} -
  \log{V}}{\log{N}^2}}{a^2 = log(N) - log(V) / log(N)^2} \deqn{\log{V_0} =
  \frac{\log{V}}{\sqrt{1 - \frac{\log{V}}{\log{N}}^2}}}{log(V0) = log(V) /
  sqrt(1 - (log(V) / log(N)^2))} The measure was derived from a formula by
  Mueller (1969, as cited in Maas, 1972). \eqn{\log{}_{e}{V_0}} is equivalent
  to \eqn{\log{V_0}}, only with \eqn{e} as the base for the logarithms. Also
  calculated are \eqn{a}, \eqn{\log{V_0}} (both not the same as before) and
  \eqn{V'} as measures of relative vocabulary growth while the text
  progresses. To calculate these measures, the first half of the text and the
  full text will be examined (see Maas, 1972, p. 67 ff. for details).  Note:
  for the current method (for a dfm) there is no computation on separate
  halves of the text.}
  
  }
}
\note{
This implements only the static measures of lexical diversity, not more
  complex measures based on windows of text such as the Mean Segmental
  Type-Token Ratio, the Moving-Average Type-Token Ratio (Covington & McFall,
  2010), the MLTD or MLTD-MA (Moving-Average Measure of Textual Lexical
  Diversity) proposed by McCarthy & Jarvis (2010) or Jarvis (no year), or the
  HD-D version of vocd-D (see McCarthy & Jarvis, 2007).  These are available
  from the package \pkg{korRpus}.
}
\examples{
mydfm <- dfm(corpus_subset(data_corpus_inaugural, Year > 1980), verbose = FALSE)
(result <- textstat_lexdiv(mydfm, c("CTTR", "TTR", "U")))
cor(textstat_lexdiv(mydfm, "all")[,-1])

}
\references{
Covington, M.A. & McFall, J.D. (2010). "Cutting the Gordian Knot: 
  The Moving-Average Type-Token Ratio (MATTR)". \emph{Journal of Quantitative 
  Linguistics} 17(2), 94--100.
  
  Herdan, Gustav. 1955. "A New Derivation and Interpretation of Yule's
  'Characteristic' \emph{K}." \emph{Zeitschrift für angewandte Mathematik und
  Physik} 6(4): 332--34.
  
  Maas, H.-D., (1972). "\"Uber den Zusammenhang zwischen Wortschatzumfang und 
  L\"ange eines Textes". \emph{Zeitschrift f\"ur Literaturwissenschaft und 
  Linguistik} 2(8), 73--96.
  
  McCarthy, P.M. & Jarvis, S. (2007). "vocd: A theoretical and empirical 
  evaluation". \emph{Language Testing} 24(4), 459--488.
  
  McCarthy, P.M. & Jarvis, S. (2010). "MTLD, vocd-D, and HD-D: A validation 
  study of sophisticated approaches to lexical diversity assessment". 
  \emph{Behaviour Research Methods} 42(2), 381--392.
  
  Michalke, Meik.  (2014) \emph{koRpus: An R Package for Text Analysis}. 
  Version 0.05-5. \url{http://reaktanz.de/?c=hacking&s=koRpus}
  
  Simpson, Edward H. 1949. "Measurement of Diversity." \emph{Nature} 163: 688.
  
  Tweedie. F.J. & Baayen, R.H. (1998). "How Variable May a Constant Be? 
  Measures of Lexical Richness in Perspective". \emph{Computers and the 
  Humanities} 32(5), 323--352.
}
\author{
Kenneth Benoit, adapted from the S4 class implementation written by 
  Meik Michalke in the \pkg{koRpus} package.
}
